/*
 * (C) Copyright 2003
 * Tait Electronics Limited, Christchurch, New Zealand
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* 
 * Use RFC 1256 to get the IP address of our local router. Use this address
 * to then derive our own IP address based on masks configured and a value 
 * read from the digital I/O pins
 */


#include <common.h>
#include <command.h>
#include <net.h>


/* globals */
int NetDiscoverCount;
IPaddr_t NetRouterAddr;

/* prototypes */
static void SolicitSend(void);
static void DiscoverHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len);


/* retry mechanism - the RFC allows us to make 3 goes at a Solicit */
static void
DiscoverTimeout (void)
{
	if (NetDiscoverCount-- > 0) {
		NetSetTimeout (3 * CFG_HZ, DiscoverTimeout);
		NetSetHandler (DiscoverHandler);
		SolicitSend();
	} else {
		eth_halt();
		NetState = NETLOOP_FAIL;	/* we did not get the reply */
	}
}

/* Handle expected response, parse address info if a Router Advertisement */
static void
DiscoverHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len)
{
	volatile IP_t *ip = (volatile IP_t *)pkt;
	volatile ushort *s;
	long pref, newpref;
	int numaddr, entrysize, length, i;

	/* check that we are looking at a valid bunch of data */
	length = ntohs(ip->ip_len);
	if (length < IP_HDR_SIZE + 4) {	/* want at least 1 router address */
		return;
	}

	s = &ip->udp_src;		/* ICMP starts here */
	/* check that we really got a Router Advertizement back (ICMP type 9) */
	if (ntohs(s[0]) != 0x0900) {	/* Advertizement code */
		return;
	}
		
	/* check that the entry size corresponds to the version of the protocol we support */
	entrysize = ntohs(s[2]) & 0xff;
	if (entrysize != 2) {
		return;
	}

	/* parse the response, keeping the IP address with the highest preference */
	pref = 0x80000001;		/* ensures we ignore preference of 0x80000000 */
	numaddr = ntohs(s[2]) >> 8;
	entrysize = 0;			/* reuse a variable */
	for (i = 0; i < numaddr; i++) {
		newpref = NetReadLong((ulong *)&s[(i * 4) + 6]);
		if (newpref > pref) {
			pref = newpref;
			NetRouterAddr = NetReadIP((void *)&s[(i * 4) + 4]);
			entrysize++;
		}
	}

	/* last ditch, use the router source address */
	if (!entrysize) {
		NetRouterAddr = NetReadIP((void *)&ip->ip_src); 
	}

	NetState = NETLOOP_SUCCESS;
}


/* initiate dialogue by sending a Router Solicit ICMP message */
static void 
SolicitSend(void)
{
	volatile IP_t *ip;
	volatile ushort *s;
	IPaddr_t SolicitIP = htonl(0xe0000002);	/* special multicast address */
	IPaddr_t DummyIP = htonl(0);
	

	ip = (volatile IP_t *)(NetTxPacket + ETHER_HDR_SIZE);

	/*
	 *	Construct an IP and ICMP header.
	 */
	ip->ip_hl_v  = 0x45;		/* IP_HDR_SIZE / 4 (not including UDP) */
	ip->ip_tos   = 0;
	ip->ip_len   = htons(IP_HDR_SIZE);
	ip->ip_id    = htons(NetIPID++);
	ip->ip_off   = htons(0x4000);	/* No fragmentation */
	ip->ip_ttl   = 255;
	ip->ip_p     = 0x01;		/* ICMP */
	ip->ip_sum   = 0;
	NetCopyIP((void*)&ip->ip_src, &DummyIP);	/* already in network byte order */
	NetCopyIP((void*)&ip->ip_dst, &SolicitIP);	/* - "" - */
	ip->ip_sum   = ~NetCksum((uchar *)ip, IP_HDR_SIZE_NO_UDP / 2);

	s = &ip->udp_src;		/* ICMP starts here */
	s[0] = htons(0x0a00);		/* Advertizement Solicit, code */
	s[1] = 0;			/* checksum */
	s[2] = 0;			/* reserved */
	s[3] = 0;
	s[1] = ~NetCksum((uchar *)s, 8/2);


	NetSetEther(NetTxPacket, NetBcastAddr, PROT_IP);
	(void) eth_send(NetTxPacket, ETHER_HDR_SIZE + IP_HDR_SIZE);

}

/* Start the whole dicovery mechanism off */
void DiscoverStart(void)
{
	NetSetTimeout (3 * CFG_HZ, DiscoverTimeout);
	NetSetHandler (DiscoverHandler);
	NetDiscoverCount = 2;		/* number of timeouts - not number of tries!! */

	SolicitSend();
}
