/*
 * (C) Copyright 2000-2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <linux/mtd/mtd.h>
#include <config.h>
#include "fw_keys.h"

typedef unsigned char uchar;

typedef struct keysdev_s {
	uchar devname[16];		/* Device name */
	ulong devoff;			/* Device offset */
	ulong keys_size;		/* keystorage size */
	ulong erase_size;		/* device erase size */
} keysdev_t;

static keysdev_t keysdevices[2];
static int curdev;

#define DEVNAME(i)    keysdevices[(i)].devname
#define DEVOFFSET(i)  keysdevices[(i)].devoff
#define KEYSSIZE(i)   keysdevices[(i)].keys_size
#define DEVESIZE(i)   keysdevices[(i)].erase_size

#define CFG_KEYS_SIZE KEYSSIZE(curdev)

#define KEYS_SIZE     getkeyssize()

typedef struct keystore_s {
	ulong crc;			/* CRC32 over data bytes    */
	uchar flags;			/* active or obsolete */
	uchar *data;
} keys_t;

static keys_t keystore;

static int HaveRedundKeys = 0;

static uchar active_flag = 1;
static uchar obsolete_flag = 0;


#define XMK_STR(x)	#x
#define MK_STR(x)	XMK_STR(x)

static uchar default_keystore[] = {
	"\0"
	"\0"			/* Termimate keys_t data with 2 NULs */
};

static int flash_io (int mode);
static uchar *keysmatch (uchar * s1, uchar * s2);
static int parse_config (void);

static inline ulong getkeyssize (void)
{
	ulong rc = CFG_KEYS_SIZE - sizeof (long);

	if (HaveRedundKeys)
		rc -= sizeof (char);
	return rc;
}

/* Print to stdout */
static void print_debug(const char *format, ...)
{
  va_list args;

#ifdef DEBUG_ENABED
  va_start(args, format);

  vprintf(format, args);

  va_end(args);
#endif
}

/* Print to stderr */
static void print_error(const char *format, ...)
{
  va_list args;

#ifdef DEBUG_ENABED
  va_start(args, format);

  vfprintf(stderr, format, args);

  va_end(args);
#endif
}

/*
 * Search the keystore for a variable.
 * Return the value, if found, or NULL, if not found.
 */
unsigned char *fw_getkeys (unsigned char *name)
{
	uchar *keys, *nxt;

	if (fw_keys_init ())
		return (NULL);

	for (keys = keystore.data; *keys; keys = nxt + 1) {
		uchar *val;

		for (nxt = keys; *nxt; ++nxt) {
			if (nxt >= &keystore.data[KEYS_SIZE]) {
				print_error("## Error: keystore not terminated\n");
				return (NULL);
			}
		}
		val = keysmatch (name, keys);
		if (!val)
			continue;
		return (val);
	}
	return (NULL);
}

/*
 * Print the key fields into the given file by converting to the TAIT
 * configuration format.
 */
static void write_key_file(FILE *file, int flag, char *name, char *value)
{
#define KEY_FIELD_SEPARATOR      0x20

  char buff[BUFLEN];
  int i;
  char *field;
  int len = strlen(value);
  char val_str[len];

  strncpy(val_str, value, len);
  val_str[len] = '\0';

  /* version and active keyset information */
  if (strcmp(name, "slot_version") == 0)
  {
    fprintf(file, "version%c\"%s\"\n", (flag) ? '=' : KEY_FIELD_SEPARATOR, value);
    return;
  }
  if (strcmp(name, "slot_keyset") == 0)
  {
    fprintf(file, "keyset%c\"%s\"\n", (flag) ? '=' : KEY_FIELD_SEPARATOR, value);
    return;
  }

  /* Get the slot id number */
  len = strlen(name);
  if (len < (KEY_SLOT_PREFIX_LEN+1) || len > (KEY_SLOT_PREFIX_LEN+2))
    return;

  len -= KEY_SLOT_PREFIX_LEN;
  strncpy(buff, name+KEY_SLOT_PREFIX_LEN, len);
  buff[len++] = (flag) ? '=' : KEY_FIELD_SEPARATOR;

  /* Get the version or key fields: CKR, keyset, algorithm id, key id, key value */
  field = strtok(val_str, " \t\n");
  if (!field)
    return;

  i = strlen(field);
  strncpy(buff+len, field, i);
  len += i;

  field = strtok(NULL, " \t\n");
  for (; field != NULL;) {
    buff[len++] = KEY_FIELD_SEPARATOR;
    i = strlen(field);
    strncpy(buff+len, field, i);
    len += i;

    field = strtok(NULL, " \t\n");
  }
  buff[len] = '\0';

  fprintf(file, "%s\n", buff);
}

/*
 * Print the current definition of one, or more, or all
 * keystore variables
 */
void fw_printkeys (int argc, char *argv[], FILE *file, int flag)
{
	uchar *keys, *nxt;
	int i;
	uchar *name;
	uchar *val = NULL;

	if (fw_keys_init ())
		return;

	if (argc == 1) {		/* Print all keys variables  */
		for (keys = keystore.data; *keys; keys = nxt + 1) {
			for (nxt = keys; *nxt; ++nxt) {
			  if (nxt >= &keystore.data[KEYS_SIZE]) {
			    print_error("## Error: keystore not terminated\n");
			    return;
			  }
			}
			val = strchr(keys, '=');
			if (val) {
				*val++ = '\0';
			        name = keys;
				if (file != stdout)
				  write_key_file(file, flag, name, val);
				else
				  fprintf(file, "%s%s%s\n", name, (flag) ? "=" : " ", val);
			}
		}
		return;
	}

	for (i = 1; i < argc; ++i) {	/* print single keys variables   */
		uchar *name = argv[i];
		uchar *val = NULL;

		for (keys = keystore.data; *keys; keys = nxt + 1) {

			for (nxt = keys; *nxt; ++nxt) {
				if (nxt >= &keystore.data[KEYS_SIZE]) {
					print_error("## Error: keystore not terminated\n");
					return;
				}
			}
			val = keysmatch (name, keys);
			if (val) {
				if (file != stdout)
				  write_key_file(file, flag, name, val);
				else
				  fprintf(file, "%s%s%s\n", name, (flag) ? "=" : " ", val);
				break;
			}
		}
		if (!val)
			print_error("## Error: \"%s\" not defined\n", name);
	}
}

/*
 * Reset all keys. 
 * NOTE: use fw_savekeys() after calling this function to remove from the flash.
 */
void fw_reset_keys (void)
{
  if (keystore.data)
    memcpy(keystore.data, &default_keystore, sizeof(default_keystore));
}

/*
 * Set new key. Returns errno style error codes:
 * 0	  - OK
 * EINVAL - need at least 1 argument
 *
 */
int fw_set_new_keys (int argc, char *argv[])
{
	int i, len;
	uchar *keys, *nxt;
	uchar *oldval = NULL;
	uchar *name;

	if (argc < 2) {
		return (EINVAL);
	}

	name = argv[1];

	/*
	 * search if variable with this name already exists
	 */
	for (nxt = keys = keystore.data; *keys; keys = nxt + 1) {
		for (nxt = keys; *nxt; ++nxt) {
			if (nxt >= &keystore.data[KEYS_SIZE]) {
				print_error("## Error: keystore not terminated\n");
				return (EINVAL);
			}
		}
	}

	/* add valid key only */
	if (argc < 3)
		return (0);

	/*
	 * Append new definition at the end
	 */
	for (keys = keystore.data; *keys || *(keys + 1); ++keys);
	if (keys > keystore.data)
		++keys;
	/*
	 * Overflow when:
	 * "name" + "=" + "val" +"\0\0"  > CFG_KEYS_SIZE - (keys-keystore)
	 */
	len = strlen (name) + 2;
	/* add '=' for first arg, ' ' for all others */
	for (i = 2; i < argc; ++i) {
		len += strlen (argv[i]) + 1;
	}
	if (len > (&keystore.data[KEYS_SIZE] - keys)) {
		print_error("Error: keystore overflow, \"%s\" deleted\n",
			    name);
		return (-1);
	}
	while ((*keys = *name++) != '\0')
		keys++;
	for (i = 2; i < argc; ++i) {
		uchar *val = argv[i];

		*keys = (i == 2) ? '=' : ' ';
		while ((*++keys = *val++) != '\0');
	}

	/* end is marked with double '\0' */
	*++keys = '\0';

	return (0);
}

/*
 * Deletes or sets keystore variables. Returns errno style error codes:
 * 0	  - OK
 * EINVAL - need at least 1 argument
 * EROFS  - certain variables ("ethaddr", "serial#") cannot be
 *	    modified or deleted
 *
 */
int fw_setkeys (int argc, char *argv[])
{
	int i, len;
	uchar *keys, *nxt;
	uchar *oldval = NULL;
	uchar *name;

	if (argc < 2) {
		return (EINVAL);
	}

	name = argv[1];

	/*
	 * search if variable with this name already exists
	 */
	for (nxt = keys = keystore.data; *keys; keys = nxt + 1) {
		for (nxt = keys; *nxt; ++nxt) {
			if (nxt >= &keystore.data[KEYS_SIZE]) {
				print_error("## Error: keystore not terminated\n");
				return (EINVAL);
			}
		}
		if ((oldval = keysmatch (name, keys)) != NULL)
			break;
	}

	/*
	 * Delete any existing definition
	 */
	if (oldval) {
		/*
		 * Ethernet Address and serial# can be set only once
		 */
		if ((strcmp (name, "ethaddr") == 0) ||
			(strcmp (name, "serial#") == 0)) {
			print_error("Can't overwrite \"%s\"\n", name);
			return (EROFS);
		}

		if (*++nxt == '\0') {
			*keys = '\0';
		} else {
			for (;;) {
				*keys = *nxt++;
				if ((*keys == '\0') && (*nxt == '\0'))
					break;
				++keys;
			}
		}
		*++keys = '\0';
	}

	/* Delete only ? */
	if (argc < 3)
		return (0);

	/*
	 * Append new definition at the end
	 */
	for (keys = keystore.data; *keys || *(keys + 1); ++keys);
	if (keys > keystore.data)
		++keys;
	/*
	 * Overflow when:
	 * "name" + "=" + "val" +"\0\0"  > CFG_KEYS_SIZE - (keys-keystore)
	 */
	len = strlen (name) + 2;
	/* add '=' for first arg, ' ' for all others */
	for (i = 2; i < argc; ++i) {
		len += strlen (argv[i]) + 1;
	}
	if (len > (&keystore.data[KEYS_SIZE] - keys)) {
		print_error("Error: keystore overflow, \"%s\" deleted\n",
			    name);
		return (-1);
	}
	while ((*keys = *name++) != '\0')
		keys++;
	for (i = 2; i < argc; ++i) {
		uchar *val = argv[i];

		*keys = (i == 2) ? '=' : ' ';
		while ((*++keys = *val++) != '\0');
	}

	/* end is marked with double '\0' */
	*++keys = '\0';

	return (0);

}


/* function split out so a single save can be used for a whole file full of data */
int fw_savekeys (void)
{

	/* Update CRC */
	keystore.crc = crc32 (0, keystore.data, KEYS_SIZE);
	
	/* write keystore back to flash */
	if (flash_io (O_RDWR)) {
		print_error("Error: can't write fw_keys to flash\n");
		return (-1);
	}

	return (0);
}


/* erase both (if redundant) sectors of the keystore */
int fw_erase (int argc, char *argv[])
{
	erase_info_t erase;
	int fd, fdr, otherdev;
	

	if (fw_keys_init ())
		return -1;

	print_debug("Erasing all keys and backups\n");
	
	if ((fd = open (DEVNAME (curdev), O_RDWR)) < 0) {
		print_error("Can't open %s: %s\n",
			DEVNAME (curdev), strerror (errno));
		return (-1);
	}
	erase.length = DEVESIZE (curdev);
	erase.start = DEVOFFSET (curdev);
	ioctl (fd, MEMUNLOCK, &erase);
	if (ioctl (fd, MEMERASE, &erase) != 0) {
		print_error("MTD erase error on %s: %s\n",
			    DEVNAME (otherdev),
			    strerror (errno));
		return (-1);
	}

	if (HaveRedundKeys) {
		otherdev = !curdev;
		if ((fdr = open (DEVNAME (otherdev), O_RDWR)) < 0) {
			print_error("Can't open %s: %s\n",
				    DEVNAME (otherdev),
				    strerror (errno));
			return (-1);
		}
		erase.length = DEVESIZE (otherdev);
		erase.start = DEVOFFSET (otherdev);
		ioctl (fdr, MEMUNLOCK, &erase);

		if (ioctl (fdr, MEMERASE, &erase) != 0) {
			print_error("MTD erase error on %s: %s\n",
				    DEVNAME (otherdev),
				    strerror (errno));
			return (-1);
		}
	}
	print_debug("Done\n");
	return(0);
}

static int flash_io (int mode)
{
	int fd, fdr, rc, otherdev, len, resid;
	erase_info_t erase;
	char *data = NULL;

	if ((fd = open (DEVNAME (curdev), mode)) < 0) {
		print_error("Can't open %s: %s\n",
			    DEVNAME (curdev), strerror (errno));
		return (-1);
	}

	len = sizeof (keystore.crc);
	if (HaveRedundKeys) {
		len += sizeof (keystore.flags);
	}

	if (mode == O_RDWR) {
		if (HaveRedundKeys) {
			/* switch to next partition for writing */
			otherdev = !curdev;
			if ((fdr = open (DEVNAME (otherdev), mode)) < 0) {
				print_error("Can't open %s: %s\n",
					    DEVNAME (otherdev),
					    strerror (errno));
				return (-1);
			}
		} else {
			otherdev = curdev;
			fdr = fd;
		}
		print_debug("Unlocking flash...\n");
		erase.length = DEVESIZE (otherdev);
		erase.start = DEVOFFSET (otherdev);
		ioctl (fdr, MEMUNLOCK, &erase);

		if (HaveRedundKeys) {
			erase.length = DEVESIZE (curdev);
			erase.start = DEVOFFSET (curdev);
			ioctl (fd, MEMUNLOCK, &erase);
			keystore.flags = active_flag;
		}

		print_debug("Done\n");
		resid = DEVESIZE (otherdev) - CFG_KEYS_SIZE;
		if (resid) {
			if ((data = malloc (resid)) == NULL) {
				print_error("Cannot malloc %d bytes: %s\n",
					    resid,
					    strerror (errno));
				return (-1);
			}
			if (lseek (fdr, DEVOFFSET (otherdev) + CFG_KEYS_SIZE, SEEK_SET)
				== -1) {
				print_error("seek error on %s: %s\n",
					    DEVNAME (otherdev),
					    strerror (errno));
				return (-1);
			}
			if ((rc = read (fdr, data, resid)) != resid) {
				print_error("read error on %s: %s\n",
					    DEVNAME (otherdev),
					    strerror (errno));
				return (-1);
			}
		}

		print_debug("Erasing old keystore...\n");

		erase.length = DEVESIZE (otherdev);
		erase.start = DEVOFFSET (otherdev);
		if (ioctl (fdr, MEMERASE, &erase) != 0) {
			print_error("MTD erase error on %s: %s\n",
				    DEVNAME (otherdev),
				    strerror (errno));
			return (-1);
		}

		print_debug("Done\n");

		print_debug("Writing keystore to %s...\n", DEVNAME (otherdev));
		if (lseek (fdr, DEVOFFSET (otherdev), SEEK_SET) == -1) {
			print_error("seek error on %s: %s\n",
				    DEVNAME (otherdev), strerror (errno));
			return (-1);
		}
		if (write (fdr, &keystore, len) != len) {
			print_error("CRC write error on %s: %s\n",
				    DEVNAME (otherdev), strerror (errno));
			return (-1);
		}
		if (write (fdr, keystore.data, KEYS_SIZE) != KEYS_SIZE) {
			print_error("Write error on %s: %s\n",
				    DEVNAME (otherdev), strerror (errno));
			return (-1);
		}
		if (resid) {
			if (write (fdr, data, resid) != resid) {
				print_error("write error on %s: %s\n",
					    DEVNAME (curdev), strerror (errno));
				return (-1);
			}
			free (data);
		}
		if (HaveRedundKeys) {
			/* change flag on current active keys partition */
			if (lseek (fd, DEVOFFSET (curdev) + sizeof (ulong), SEEK_SET)
				== -1) {
				print_error("seek error on %s: %s\n",
					    DEVNAME (curdev), strerror (errno));
				return (-1);
			}
			if (write (fd, &obsolete_flag, sizeof (obsolete_flag)) !=
				sizeof (obsolete_flag)) {
				print_error("Write error on %s: %s\n",
					    DEVNAME (curdev), strerror (errno));
				return (-1);
			}
		}

		print_debug("Done\n");
		print_debug("Locking ...\n");
		erase.length = DEVESIZE (otherdev);
		erase.start = DEVOFFSET (otherdev);
		ioctl (fdr, MEMLOCK, &erase);
		if (HaveRedundKeys) {
			erase.length = DEVESIZE (curdev);
			erase.start = DEVOFFSET (curdev);
			ioctl (fd, MEMLOCK, &erase);
			if (close (fdr)) {
				print_error("I/O error on %s: %s\n",
					    DEVNAME (otherdev),
					    strerror (errno));
				return (-1);
			}
		}
		print_debug("Done\n");
	} else {

		if (lseek (fd, DEVOFFSET (curdev), SEEK_SET) == -1) {
			print_error("seek error on %s: %s\n",
				    DEVNAME (curdev), strerror (errno));
			return (-1);
		}
		if (read (fd, &keystore, len) != len) {
			print_error("CRC read error on %s: %s\n",
				    DEVNAME (curdev), strerror (errno));
			return (-1);
		}
		if ((rc = read (fd, keystore.data, KEYS_SIZE)) != KEYS_SIZE) {
		        print_error("Read error on %s: %s\n",
				    DEVNAME (curdev), strerror (errno));
			return (-1);
		}
	}

	if (close (fd)) {
		print_error("I/O error on %s: %s\n",
			    DEVNAME (curdev), strerror (errno));
		return (-1);
	}

	/* everything ok */
	return (0);
}

/*
 * s1 is either a simple 'name', or a 'name=value' pair.
 * s2 is a 'name=value' pair.
 * If the names match, return the value of s2, else NULL.
 */

static uchar *keysmatch (uchar * s1, uchar * s2)
{

	while (*s1 == *s2++)
		if (*s1++ == '=')
			return (s2);
	if (*s1 == '\0' && *(s2 - 1) == '=')
		return (s2);
	return (NULL);
}

/*
 * Prevent confusion if running from erased flash memory
 */
int fw_keys_init (void)
{
	int crc1, crc1_ok;
	uchar *addr1;

	int crc2, crc2_ok;
	uchar flag1, flag2, *addr2;

	if (parse_config ())		/* should fill keysdevices */
		return 1;

	if ((addr1 = calloc (1, KEYS_SIZE)) == NULL) {
		print_error("Not enough memory for keystore (%ld bytes)\n",
			    KEYS_SIZE);
		return (errno);
	}

	/* read keystore from FLASH to local buffer */
	keystore.data = addr1;
	curdev = 0;
	if (flash_io (O_RDONLY)) {
		return (errno);
	}

	crc1_ok = ((crc1 = crc32 (0, keystore.data, KEYS_SIZE))
			   == keystore.crc);
	
	if (!HaveRedundKeys) {
		if (!crc1_ok) {
			print_error("Warning: Bad CRC, keystore is empty\n");
			memcpy(keystore.data, &default_keystore, sizeof(default_keystore));
		}
	} else {
		flag1 = keystore.flags;

		curdev = 1;
		if ((addr2 = calloc (1, KEYS_SIZE)) == NULL) {
			print_error("Not enough memory for keystore (%ld bytes)\n",
				    KEYS_SIZE);
			return (errno);
		}
		keystore.data = addr2;

		if (flash_io (O_RDONLY)) {
			return (errno);
		}

		crc2_ok = ((crc2 = crc32 (0, keystore.data, KEYS_SIZE))
				   == keystore.crc);
		flag2 = keystore.flags;

		if (crc1_ok && !crc2_ok) {
			keystore.data = addr1;
			keystore.flags = flag1;
			keystore.crc = crc1;
			curdev = 0;
			free (addr2);
		} else if (!crc1_ok && crc2_ok) {
			keystore.data = addr2;
			keystore.flags = flag2;
			keystore.crc = crc2;
			curdev = 1;
			free (addr1);
		} else if (!crc1_ok && !crc2_ok) {
			print_error("Warning: Bad CRC, keystore is empty\n");
			keystore.data = addr1;
			keystore.flags = flag1;
			keystore.crc = crc1;
			curdev = 0;
			free (addr2);
			memcpy(keystore.data, &default_keystore, sizeof(default_keystore));
		} else if (flag1 == active_flag && flag2 == obsolete_flag) {
			keystore.data = addr1;
			keystore.flags = flag1;
			keystore.crc = crc1;
			curdev = 0;
			free (addr2);
		} else if (flag1 == obsolete_flag && flag2 == active_flag) {
			keystore.data = addr2;
			keystore.flags = flag2;
			keystore.crc = crc2;
			curdev = 1;
			free (addr1);
		} else if (flag1 == flag2) {
			keystore.data = addr1;
			keystore.flags = flag1;
			keystore.crc = crc1;
			curdev = 0;
			free (addr2);
		} else if (flag1 == 0xFF) {
			keystore.data = addr1;
			keystore.flags = flag1;
			keystore.crc = crc1;
			curdev = 0;
			free (addr2);
		} else if (flag2 == 0xFF) {
			keystore.data = addr2;
			keystore.flags = flag2;
			keystore.crc = crc2;
			curdev = 1;
			free (addr1);
		}
	}
	return (0);
}


static int parse_config ()
{
	struct stat st;

	strcpy (DEVNAME (0), DEVICE1_NAME);
	DEVOFFSET (0) = DEVICE1_OFFSET;
	KEYSSIZE (0) = KEYS1_SIZE;
	DEVESIZE (0) = DEVICE1_ESIZE;
#ifdef HAVE_REDUND
	strcpy (DEVNAME (1), DEVICE2_NAME);
	DEVOFFSET (1) = DEVICE2_OFFSET;
	KEYSSIZE (1) = KEYS2_SIZE;
	DEVESIZE (1) = DEVICE2_ESIZE;
	HaveRedundKeys = 1;
#endif
	if (stat (DEVNAME (0), &st)) {
		print_error("Cannot access MTD device %s: %s\n",
			    DEVNAME (0), strerror (errno));
		return 1;
	}

	if (HaveRedundKeys && stat (DEVNAME (1), &st)) {
		print_error("Cannot access MTD device %s: %s\n",
			    DEVNAME (2), strerror (errno));
		return 1;
	}
	return 0;
}

