/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Command line user interface to firmware (=U-Boot) keysironment.
 *
 * Implements:
 *	fw_printkeys [ name ... ]
 *		- prints the values of the keystore variables
 *		  "name", or the whole keysironment if no names are
 *		  specified
 *	fw_setkeys name [ value ... ]
 *		- If a name without any values is given, the variable
 *		  with this name is deleted from the keystore;
 *		  otherwise, all "value" arguments are concatenated,
 *		  separated by sinlge blank characters, and the
 *		  resulting string is assigned to the keystore
 *		  variable "name"
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>

#include "fw_keys.h"

#define	CMD_PRINTKEYS	"keysread"
#define CMD_SETKEYS	"keysedit"
#define CMD_ERASE       "keyserase"

#define FLEN 100
#define CFG_MAXARGS 10

char filename[FLEN+1];
int havefile = 0;
char buf[BUFLEN+1];
int equals = 0;


/* pull out, validate and set flags for all command line options and parameters */
static int parseArguments(int argc, char **argv)
{
    int c;
    static struct option long_options[] = {
        {"help", 0, NULL, 'h'},
	{"file", 1, NULL, 'f'},
	{"equals", 0, NULL, 'e'},
        {NULL, 0, NULL, 0}
    };

    while (1)
    {
        c = getopt_long(argc, argv, "ef:h", long_options, NULL);
        if (c == -1)
            break;

        switch (c)
        {
            case 'h':
                printf("%s: Key store maintenance program\n"
                        "usage: %s [-ehf <filename>]\n"
                        "\n"
			"-e, --equals           Use equals sign to separate name from value\n"
			"-f, --file <filename>  Get/put data from/to file\n"
                        "-h, --help             This help screen\n",
                        argv[0], argv[0]);
                exit(0);
	    case 'e':
		equals = 1;
		break;
	    case 'f':
		if(optarg) {
		    strncpy(filename, optarg, FLEN);
		    havefile = 1;
		}
	        else
        	    printf("Unknown arg \"%s\", use %s --help\n", optarg, argv[0]);

		break;
            case '?':
                printf("Unknown option \"%s\"", argv[optind - 1]);
                break;
            default:
                break;
        }
    }
    return (optind);
}

/* chop up input line into arguments */
int parse_line (char *line, char *argv[])
{
	int nargs = 1;		/* argv[0] not used */

	while (nargs < CFG_MAXARGS) {
		/* skip any white space (including Line Feed) */
		while ((*line == ' ') || (*line == '\t') || (*line == '\n')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
			return (nargs);
		}
		
		argv[nargs++] = line;	/* begin of argument string	*/
		/* find end of string */
		while (*line && (*line != ' ') && (*line != '\t') && (*line != '\n') && ((equals)?(*line != '='):~0)) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
			return (nargs);
		}

		*line++ = '\0';		/* terminate current arg	 */
	}

	return (nargs);
}


int
main(int argc, char *argv[])
{
	char *p;
	char *cmdname = *argv;
	int opt, i, len;
	FILE *file;
	
        opt = parseArguments(argc, argv);

	/* shuffle parameters down so the rest of the code doesn't have to worry about options */
	/* leave argv[0] as it is!! */
	for (i = 1; i < argc; i++) {
	    argv[i] = argv[i + opt - 1];
	}
	argc = argc - opt + 1;

	if ((p = strrchr (cmdname, '/')) != NULL) {
		cmdname = p + 1;
	}

	if (strcmp(cmdname, CMD_PRINTKEYS) == 0) {
		if (havefile) {
		        file = fopen(filename, "w");
			if (!file) {
#ifdef DEBUG_ENABLED
			        printf ("Can't open file %s - %s\n", filename, strerror(errno));
#endif
				return (EXIT_FAILURE);
			}
			
			fw_printkeys (argc, argv, file, equals);
			fclose(file);			
		} else {
			fw_printkeys (argc, argv, stdout, equals);
		}
		return (EXIT_SUCCESS);

	} else if (strcmp(cmdname, CMD_SETKEYS) == 0) {
		if (fw_keys_init ())
			return (EXIT_FAILURE);

		if (havefile) {
		        file = fopen(filename, "r");
			if (!file) {
#ifdef DEBUG_ENABLED
			        printf ("Can't open file %s - %s\n", filename, strerror(errno));
#endif
				return (EXIT_FAILURE);
			}

			/* reset keys stored */
			fw_reset_keys();

			/* attach the key slot name prefix */
			strncpy(buf, KEY_SLOT_PREFIX, KEY_SLOT_PREFIX_LEN);
			len = BUFLEN - KEY_SLOT_PREFIX_LEN;
			while (fgets(buf+KEY_SLOT_PREFIX_LEN, len, file)) {
				argc = parse_line(buf, argv);
				if (fw_set_new_keys (argc, argv) != 0)
					return (EXIT_FAILURE);
				strncpy(buf, KEY_SLOT_PREFIX, KEY_SLOT_PREFIX_LEN);
			} 
			
			fclose(file);			
			if (fw_savekeys() !=0)
				return (EXIT_FAILURE);

		} else {
#ifdef DEBUG_ENABLED
			printf("argc %d, argv1 %s, argv2 %s\n", argc, argv[1], argv[2]);
#endif
			if (fw_setkeys (argc, argv) != 0)
				return (EXIT_FAILURE);
			if (fw_savekeys() !=0)
				return (EXIT_FAILURE);
		}
		return (EXIT_SUCCESS);
	} else if (strcmp(cmdname, CMD_ERASE) == 0) {
	        if (fw_erase (argc, argv) != 0)
	               return (EXIT_FAILURE);

		return (EXIT_SUCCESS);
	}
#ifdef DEBUG_ENABLED
	fprintf (stderr,
		"Identity crisis - may be called as `" CMD_PRINTKEYS
		"' or as `" CMD_SETKEYS "' or as `" CMD_ERASE "' but not as `%s'\n",
		cmdname);
#endif
	return (EXIT_FAILURE);
}
